# Copyright (c) 2018 ADLINK Technology Inc.
#
# See the NOTICE file(s) distributed with this work for additional
# information regarding copyright ownership.
#
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0, or the Apache License, Version 2.0
# which is available at https://www.apache.org/licenses/LICENSE-2.0.
#
# SPDX-License-Identifier: EPL-2.0 OR Apache-2.0
#
# Contributors: Gabriele Baldoni, ADLINK Technology Inc. - Yaks API

import re
from yaks.exceptions import ValidationError
from yaks.path import Path


class Selector(object):
    def __init__(self, selector):
        self.__sel_regex = re.compile(
            '^([^?#]+)(\?([^\[()\]#]*)(\((.*)\))?)?(\#(.*))?$')
        if not self.is_valid(selector):
            raise ValidationError(
                "{} is not a valid Selector".format(selector))
        res = self.__sel_regex.match(selector)
        self.selector = selector
        self.path = res.group(1)
        self.predicate = res.group(3)
        self.properties = res.group(5)
        self.fragment = res.group(7)

    @staticmethod
    def to_selector(s):
        if isinstance(s, Selector):
            return s
        else:
            return Selector(s)

    def is_valid(self, selector):
        return self.__sel_regex.match(selector) is not None \
            and not selector.startswith('//')

    def is_absolute(self):
        if self.path.startswith('/'):
            return True
        return False

    def is_path_unique(self):
        return '*' not in self.path

    def is_prefixed_by_path(self, path):
        return self.path.startswith(str(path))

    def to_string(self):
        return self.selector

    def get_path(self):
        return Path(self.path.split('*')[0])

    def get_predicate(self):
        return self.predicate

    def get_fragment(self):
        return self.fragment

    def get_properties(self):
        return self.properties

    def dict_from_properties(self):
        data = {}
        if self.properties is None:
            return data
        uri_values = self.properties.split(';')
        for tokens in uri_values:
            vs = tokens.split('=')[1:]
            if len(vs) > 1:
                v = '='.join(vs)
            else:
                v = vs[0]
            k = tokens.split('=')[0]
            if len(k.split('.')) < 2:
                data.update({k: v})
            else:
                d = self.__dot2dict(k, v)
                data.update(d)
        return data

    def __dot2dict(self, dot_notation, value=None):
        ld = []

        tokens = dot_notation.split('.')
        n_tokens = len(tokens)
        for i in range(n_tokens, 0, -1):
            if i == n_tokens and value is not None:
                ld.append({tokens[i - 1]: value})
            else:
                ld.append({tokens[i - 1]: ld[-1]})
        return ld[-1]

    def __len__(self):
        return len(self.selector)

    def __eq__(self, second_selector):
        if isinstance(second_selector, self.__class__):
            return self.selector == second_selector.selector
        return False

    def __str__(self):
        return self.selector

    def __repr__(self):
        return self.__str__()

    def __hash__(self):
        return self.selector.__hash__()



# Copyright (c) 2018 ADLINK Technology Inc.
#
# See the NOTICE file(s) distributed with this work for additional
# information regarding copyright ownership.
#
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0, or the Apache License, Version 2.0
# which is available at https://www.apache.org/licenses/LICENSE-2.0.
#
# SPDX-License-Identifier: EPL-2.0 OR Apache-2.0
#
# Contributors: Gabriele Baldoni, ADLINK Technology Inc. - Yaks API


class ValidationError(Exception):
    def __init__(self, message, errors=0):
        super().__init__(message)
        self.errors = errors



# Copyright (c) 2018 ADLINK Technology Inc.
#
# See the NOTICE file(s) distributed with this work for additional
# information regarding copyright ownership.
#
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0, or the Apache License, Version 2.0
# which is available at https://www.apache.org/licenses/LICENSE-2.0.
#
# SPDX-License-Identifier: EPL-2.0 OR Apache-2.0
#
# Contributors: Angelo Corsaro, ADLINK Technology Inc. - Yaks API refactoring

from yaks.value import Value
from yaks.encoding import Encoding


class Admin(object):

    PREFIX = '@'
    MY_YAKS = 'local'

    def __init__(self, ws):
        self.ws = ws

    def add_frontend(self, feid, properties, yaks=MY_YAKS):
        '''
        Not supported in this version.
        '''
        path = '/{}/{}/frontend/{}'.format(Admin.PREFIX, yaks, feid)
        value = Value(properties, encoding=Encoding.PROPERTY)
        return self.ws.put(path, value, quorum=1)

    def get_frontends(self, yaks=MY_YAKS):
        '''
        Returns the list of frontends available on the Yaks
        instance with UUID **yaks**.
        '''
        s = '/{}/{}/frontend/*'.format(Admin.PREFIX, yaks)
        kvs = self.ws.get(s)
        return list(map(lambda e: (e[0].split('/')[-1], e[1].value), kvs))

    def get_frontend(self, feid, yaks=MY_YAKS):
        '''
        Returns the frontend with the front-end ID **feid**
        on the Yaks instance with UUID **yaks**.
        '''
        s = '/{}/{}/frontend/{}'.format(Admin.PREFIX, yaks, feid)
        kvs = self.ws.get(s)
        if len(kvs) > 0:
            return kvs[0][1].value
        return None

    def remove_frontend(self, feid, yaks=MY_YAKS):
        '''
        Not supported in this version.
        '''
        path = '/{}/{}/frontend/{}'.format(Admin.PREFIX, yaks, feid)
        return self.ws.remove(path, quorum=1)

    def add_backend(self, beid, properties, yaks=MY_YAKS):
        '''
        Not supported in this version.
        '''
        path = '/{}/{}/backend/{}'.format(Admin.PREFIX, yaks, beid)
        value = Value(properties, encoding=Encoding.PROPERTY)
        return self.ws.put(path, value, quorum=1)

    def get_backends(self, yaks=MY_YAKS):
        '''
        Gets the list of all available back-ends on the Yaks instance
        with UUID **yaks**.
        '''
        s = '/{}/{}/backend/*'.format(Admin.PREFIX, yaks)
        kvs = self.ws.get(s)
        return list(map(lambda e: (e[0].split('/')[-1], e[1].value), kvs))

    def get_backend(self, beid, yaks=MY_YAKS):
        '''
        Gets the  back-end with id **beid** on the Yaks instance
        with UUID **yaks**.
        '''
        s = '/{}/{}/backend/{}'.format(Admin.PREFIX, yaks, beid)
        kvs = self.ws.get(s)
        if len(kvs) > 0:
            return kvs[0][1].value
        return None

    def remove_backend(self, beid, yaks=MY_YAKS):
        '''
        Not supported in this version.
        '''
        path = '/{}/{}/backend/{}'.format(Admin.PREFIX, yaks, beid)
        return self.ws.remove(path, quorum=1)

    def add_storage(self, stid, properties, beid=None, yaks=MY_YAKS):
        '''

        Adds a storage named **stid** on the backend **beid** and with
        storage and back-end specific configuration defined through
        **properties**.
        The **properties** should always include the selector,
        e.g., *{"selector":"/demo/astore/**"}*.
        Main memory is the default backend used when **beid** is unset.

        Finally, the storage is created on the Yaks instance with UUID
        **yaks**.

        '''

        if not beid:
            beid = 'auto'
        p = '/{}/{}/backend/{}/storage/{}'.format(
            Admin.PREFIX, yaks, beid, stid)
        v = Value(properties, encoding=Encoding.PROPERTY)
        return self.ws.put(p, v, quorum=1)

    def get_storages(self, beid=None, yaks=MY_YAKS):
        '''
        Gets the list of all available storages on the Yaks instance
        with UUID **yaks**.
        '''
        if not beid:
            beid = '*'
        s = '/{}/{}/backend/{}/storage/*'.format(Admin.PREFIX, yaks, beid)
        kvs = self.ws.get(s)
        return list(map(lambda e: (e[0].split('/')[-1], e[1].value), kvs))

    def get_storage(self, stid, yaks=MY_YAKS):
        '''
        Gets the  storage with id **stid** on the Yaks instance
        with UUID **yaks**.
        '''
        s = '/{}/{}/backend/*/storage/{}'.format(Admin.PREFIX, yaks, stid)
        kvs = self.ws.get(s)
        if len(kvs) > 0:
            return kvs[0][1].value
        return None

    def remove_storage(self, stid, yaks=MY_YAKS):
        '''
        Removes the  storage with id **stid** on the Yaks instance
        with UUID **yaks**.
        '''
        s = '/{}/{}/backend/*/storage/{}'.format(Admin.PREFIX, yaks, stid)
        kvs = self.ws.get(s)
        if len(kvs) > 0:
            p = kvs[0][0]
            return self.ws.remove(p, quorum=1)
        return False

    def get_sessions(self, yaks=MY_YAKS, feid=None):
        '''
        Gets the list of all available sessions on the Yaks instance
        with UUID **yaks**.
        '''
        if not feid:
            feid = '*'
        s = '/{}/{}/frontend/{}/session/*'.format(Admin.PREFIX, yaks, feid)
        kvs = self.ws.get(s)
        return list(map(lambda e: (e[0].split('/')[-1], e[1].value), kvs))

    def close_session(self, sid, yaks=MY_YAKS):
        '''
        Not supported in this version.
        '''
        s = '/{}/{}/frontend/*/session/{}'.format(Admin.PREFIX, yaks, sid)
        kvs = self.ws.get(s)
        if len(kvs) > 0:
            p = kvs[0][0]
            return self.ws.remove(p, quorum=1)
        return False

    def get_subscriptions(self, sid, yaks=MY_YAKS):
        '''
        Gets the list of all active subscriptions on the Yaks instance
        with UUID **yaks**.
        '''
        s = '/{}/{}/frontend/*/session/{}'.format(Admin.PREFIX, yaks, sid)
        kvs = self.ws.get(s)
        return list(map(lambda e: (e[0].split('/')[-1], e[1].value), kvs))



# Copyright (c) 2018 ADLINK Technology Inc.
#
# See the NOTICE file(s) distributed with this work for additional
# information regarding copyright ownership.
#
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0, or the Apache License, Version 2.0
# which is available at https://www.apache.org/licenses/LICENSE-2.0.
#
# SPDX-License-Identifier: EPL-2.0 OR Apache-2.0
#
# Contributors: Angelo Corsaro, ADLINK Technology Inc. - Yaks API refactoring
#
#
#
# 7 6 5 4 3 2 1 0
# +-+-+-+-+-+-+-+-+ ----------------------+
# |  MESSAGE CODE |    8bit               |
# +-+-+-+-+-+-+-+-+                       |
# |X|X|X|X|X|A|S|P|    8bit               +--> Header
# +-+-+-+-+-+-+-+-+                       |
# ~   Corr. ID    ~  VLE max 64bit        |
# +---------------+                       |
# ~   Properties  ~ --> Present if P = 1  |
# +---------------+ ----------------------+
# ~     Body      ~ --> its structure depends on the message code
# +---------------+
#
#
# WIRE MESSAGE for framing on TCP/IP:
#
#  7 6 5 4 3 2 1 0
# +-+-+-+-+-+-+-+-+
# ~    Length     ~ VLE max 64bit
# +-+-+-+-+-+-+-+-+ ----------------------+
# |  MESSAGE CODE |    8bit               |
# +-+-+-+-+-+-+-+-+                       |
# |X|X|X|X|X|A|S|P|    8bit               +--> Header
# +-+-+-+-+-+-+-+-+                       |
# ~    Corr. id   ~  VLE max 64bit        |
# +---------------+                       |
# ~   Properties  ~ --> Present if P = 1  |
# +---------------+ ----------------------+
# ~     Body      ~ VL
# +---------------+
#


from papero import Property, find_property
import random
import json
import hexdump
from enum import Enum
from yaks.value import Value
from yaks.path import Path
from yaks.selector import Selector


class Message(object):
    LOGIN = 0x01
    LOGOUT = 0x02
    WORKSPACE = 0x03

    PUT = 0xA0
    UPDATE = 0xA1
    GET = 0xA2
    DELETE = 0xA3

    SUB = 0xB0
    UNSUB = 0xB1
    NOTIFY = 0xB2

    REG_EVAL = 0xC0
    UNREG_EVAL = 0xC1
    EVAL = 0xC2

    OK = 0xD0
    VALUES = 0xD1

    ERROR = 0xE0

    WSID = "wsid"
    AUTO = "auto"
    SUBID = "subid"

    def __init__(self, mid):
        self.mid = mid


class Header(Message):
    # Property Flag
    P_FLAG = 0x01
    # Incomplete Result
    # This flag is only relevant for values.
    I_FLAG = 0x02
    FLAGS_MASK = 0x01

    @staticmethod
    def has_flag(h, f):
        return h & f != 0

    def __init__(self, mid, flags=0, corr_id=None,
                 properties=None):
        super(Header, self).__init__(mid)

        if corr_id is None:
            self.corr_id = random.getrandbits(16)
        else:
            self.corr_id = corr_id

        self.flags = flags
        self.properties = properties
        if properties is not None:
            self.flags = self.flags | Header.P_FLAG

    def has_properties(self):
        return self.flags & Header.P_FLAG != 0

    def is_complete(self):
        return self.flags & Header.I_FLAG == 0


class LoginM(Header):
    def __init__(self, properties=None):
        super(LoginM, self).__init__(Message.LOGIN, properties=properties)


class LogoutM(Header):
    def __init__(self):
        super(LogoutM, self).__init__(Message.LOGOUT)


class WorkspaceM(Header):
    def __init__(self, path, properties=None):
        super(WorkspaceM, self).__init__(Message.WORKSPACE,
                                         properties=properties)
        self.path = path


class WorkspaceMessage(Header):
    def __init__(self, mid, wsid, properties=None):
        if properties is None:
            properties = []
        properties.append(Property(Message.WSID, wsid))
        super(WorkspaceMessage, self).__init__(mid, properties=properties)
        self.wsid = wsid


class PutM(WorkspaceMessage):
    def __init__(self, wsid, kvs, properties=None):
        super(PutM, self).__init__(Message.PUT, wsid, properties)
        self.kvs = kvs

    @staticmethod
    def make(kvs, properties=None):
        wsid = ''
        if properties:
            wsid = find_property(Message.WSID, properties)
        return PutM(wsid, kvs, properties)


class GetM(WorkspaceMessage):
    def __init__(self, wsid, selector, properties=None):
        super(GetM, self).__init__(Message.GET, wsid, properties)
        self.selector = selector

    @staticmethod
    def make(selector, properties=None):
        wsid = ''
        if properties:
            wsid = find_property(Message.WSID, properties)
        return GetM(wsid, selector, properties)


class UpdateM(WorkspaceMessage):
    def __init__(self, wsid, kvs, properties=None):
        super(UpdateM, self).__init__(Message.UPDATE, wsid, properties)
        self.kvs = kvs

    @staticmethod
    def make(kvs, properties=None):
        wsid = ''
        if properties:
            wsid = find_property(Message.WSID, properties)
        return UpdateM(wsid, kvs, properties)


class DeleteM(WorkspaceMessage):
    def __init__(self, wsid, path, properties=None):
        super(DeleteM, self).__init__(Message.DELETE, wsid, properties)
        self.path = path

    @staticmethod
    def make(path, properties=None):
        wsid = ''
        if properties:
            wsid = find_property(Message.WSID, properties)
        return DeleteM(wsid, path, properties)


class SubscribeM(WorkspaceMessage):
    def __init__(self, wsid, selector, properties=None):
        super(SubscribeM, self).__init__(Message.SUB, wsid, properties)
        self.selector = selector

    @staticmethod
    def make(selector, properties=None):
        wsid = ''
        if properties:
            wsid = find_property(Message.WSID, properties)
        return SubscribeM(wsid, selector, properties)


class UnsubscribeM(WorkspaceMessage):
    def __init__(self, wsid, subid, properties=None):
        super(UnsubscribeM, self).__init__(Message.UNSUB, wsid, properties)
        self.subid = subid

    @staticmethod
    def make(subid, properties=None):
        wsid = ''
        if properties:
            wsid = find_property(Message.WSID, properties)
        return UnsubscribeM(wsid, subid, properties)


class NotifyM(WorkspaceMessage):
    def __init__(self, wsid, subid, kvs, properties=None):
        super(NotifyM, self).__init__(Message.NOTIFY, wsid, properties)
        self.subid = subid
        self.kvs = kvs

    @staticmethod
    def make(subid, kvs, properties=None):
        wsid = ''
        if properties:
            wsid = find_property(Message.WSID, properties)
        return NotifyM(wsid, subid, kvs, properties)


class EvalM(WorkspaceMessage):
    def __init__(self, wsid, selector, properties=None):
        super(EvalM, self).__init__(Message.EVAL, wsid, properties)
        self.selector = selector

    @staticmethod
    def make(selector, header):
        wsid = ''
        if header.properties:
            wsid = find_property(Message.WSID, header.properties)
        m = EvalM(wsid, selector, header.properties)
        m.corr_id = header.corr_id
        return m


class RegisterEvalM(WorkspaceMessage):
    def __init__(self, wsid, path, properties=None):
        super(RegisterEvalM, self).__init__(Message.REG_EVAL, wsid, properties)
        self.path = path

    @staticmethod
    def make(path, properties=None):
        wsid = ''
        if properties:
            wsid = find_property(Message.WSID, properties)
        return RegisterEvalM(wsid, path, properties)


class UnregisterEvalM(WorkspaceMessage):
    def __init__(self, wsid, path, properties=None):
        super(UnregisterEvalM, self).__init__(
            Message.UNREG_EVAL, wsid, properties)
        self.path = path

    @staticmethod
    def make(path, properties=None):
        wsid = ''
        if properties:
            wsid = find_property(Message.WSID, properties)
        return UnregisterEvalM(wsid, path, properties)


class ValuesM(Header):
    def __init__(self, kvs):
        super(ValuesM, self).__init__(Message.VALUES)
        self.kvs = kvs

    @staticmethod
    def make(header, kvs):
        vs = ValuesM(kvs)
        vs.flags = header.flags
        vs.corr_id = header.corr_id
        return vs


class OkM(Header):
    def __init__(self):
        super(OkM, self).__init__(Message.OK)

    @staticmethod
    def make(header):
        vs = OkM()
        vs.corr_id = header.corr_id
        vs.properties = header.properties
        return vs


class ErrorM(Header):
    BAD_REQUEST = 400
    FORBIDDEN = 403
    NOT_FOUND = 404
    PRECONDITION_FAILED = 412
    INTERNAL_SERVER_ERROR = 500
    NOT_IMPLEMENTED = 501
    INSUFFICIENT_STORAGE = 507

    def __init__(self, error_code):
        super(ErrorM, self).__init__(Message.ERROR)
        self.error_code = error_code

    @staticmethod
    def make(corr_id, error_code, properties=None):
        e = ErrorM(error_code)
        e.corr_id = corr_id
        e.properties = properties
        return e



# Copyright (c) 2018 ADLINK Technology Inc.
#
# See the NOTICE file(s) distributed with this work for additional
# information regarding copyright ownership.
#
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0, or the Apache License, Version 2.0
# which is available at https://www.apache.org/licenses/LICENSE-2.0.
#
# SPDX-License-Identifier: EPL-2.0 OR Apache-2.0
#
# Contributors: Angelo Corsaro, ADLINK Technology Inc. - Yaks API refactoring

from papero import find_property, encode_sequence, decode_sequence
from papero import encode_properties, decode_properties, Property
from yaks.message import Message, Header, WorkspaceM, OkM, ErrorM, GetM, PutM
from yaks.message import UpdateM, DeleteM, SubscribeM, UnsubscribeM
from yaks.message import EvalM, RegisterEvalM, UnregisterEvalM, ValuesM
from yaks.message import NotifyM
from yaks.encoding import Encoding
from yaks.value import Value, Change
from yaks.path import Path
from yaks.selector import Selector


def encode_raw_value(buf, v):
    buf.put_string(v.raw_format)
    buf.put_bytes(v.value)


def decode_raw_value(buf):
    raw_format = buf.get_string()
    data = buf.get_bytes()
    return Value(data, encoding=Encoding.RAW, raw_format=raw_format)


def encode_string_value(buf, v):
    buf.put_string(str(v))


def decode_string_value(buf):
    v = buf.get_string()
    return Value(v, encoding=Encoding.STRING)


def encode_json_value(buf, v):
    buf.put_string(v.value)


def decode_json_value(buf):
    data = buf.get_string()
    #json.loads()
    return Value(data, encoding=Encoding.JSON)


def decode_timestamp(buf):
    ts_time = buf.get_vle()
    ts_uuid = buf.get_n_bytes(16)
    return (ts_time, ts_uuid)


def encode_property_value(buf, v):
    encode_properties(buf, v.value)


def decode_property_value(buf):
    return Value(decode_properties(buf), encoding=Encoding.PROPERTY)


def encode_value(buf, v):
    buf.put(v.encoding)
    {
        Encoding.RAW: lambda v: encode_raw_value(buf, v),
        Encoding.STRING: lambda v: encode_string_value(buf, v),
        Encoding.JSON: lambda v: encode_json_value(buf, v),
        Encoding.PROPERTY: lambda v: encode_property_value(buf, v)
    }.get(v.encoding)(v)


def decode_value(buf):
    encoding = buf.get()
    v = {
        Encoding.RAW: lambda b: decode_raw_value(b),
        Encoding.STRING: lambda b: decode_string_value(b),
        Encoding.JSON: lambda b: decode_json_value(b),
        Encoding.PROPERTY: lambda b: decode_property_value(b)
    }.get(encoding)(buf)
    return v


def decode_change(buf):
    k = chr(buf.get())
    ts = decode_timestamp(buf)
    c = Change(k, ts)
    if k in ['P', 'U']:
        v = decode_value(buf)
        c.set_value(v)
    return c


def encode_key_value(buf, kv):
    path, value = kv
    buf.put_string(str(path))
    encode_value(buf, value)


def decode_key_value(buf):
    k = buf.get_string()
    v = decode_value(buf)
    return (k, v)


def decode_key_change(buf):
    k = buf.get_string()
    v = decode_change(buf)
    return (k, v)


def encode_key_value_list(buf, kvs):
    encode_sequence(buf, kvs, encode_key_value)


def decode_key_value_list(buf):
    return decode_sequence(buf, decode_key_value)


def decode_key_change_list(buf):
    return decode_sequence(buf, decode_key_change)


def encode_header(buf, m):
    buf.put(m.mid)
    buf.put(m.flags)
    buf.put_vle(m.corr_id)
    if m.properties is not None:
        encode_properties(buf, m.properties)
    return buf


def decode_header(buf):
    mid = buf.get()
    flags = buf.get()
    corr_id = buf.get_vle()
    properties = None
    if Header.has_flag(flags, Header.P_FLAG):
        properties = decode_properties(buf)

    return Header(mid, flags, corr_id, properties)


def encode_login(buf, m):
    encode_header(buf, m)


def encode_logout(buf, m):
    encode_header(buf, m)


def encode_workspace(buf, m):
    encode_header(buf, m)
    buf.put_string(str(m.path))


def encode_put(buf, m):
    encode_header(buf, m)
    encode_key_value_list(buf, m.kvs)


def encode_get(buf, m):
    encode_header(buf, m)
    buf.put_string(str(m.selector))


def encode_update(buf, m):
    encode_header(buf, m)
    encode_key_value_list(buf, m.kvs)


def encode_delete(buf, m):
    encode_header(buf, m)
    buf.put_string(str(m.path))


def encode_sub(buf, m):
    encode_header(buf, m)
    buf.put_string(str(m.selector))


def encode_unsub(buf, m):
    encode_header(buf, m)
    buf.put_string(m.subid)


def encode_notify(buf, m):
    encode_header(buf, m)
    buf.put_string(m.subid)
    encode_key_value_list(buf, m.kvs)


def encode_reg_eval(buf, m):
    encode_header(buf, m)
    buf.put_string(str(m.path))


def encode_unreg_eval(buf, m):
    encode_header(buf, m)
    buf.put_string(str(m.path))


def encode_eval(buf, m):
    encode_header(buf, m)
    buf.put_string(str(m.selector))


def encode_values(buf, m):
    encode_header(buf, m)
    encode_key_value_list(buf, m.kvs)


def encode_ok(buf, m):
    encode_header(buf, m)


def encode_error(buf, m):
    encode_header(buf, m)
    buf.put_vle(m.error_code)


def decode_put(buf, header):
    kvs = decode_key_value_list(buf)
    return PutM.make(kvs, header.properties)


def decode_get(buf, header):
    selector = Selector(buf.get_string())
    return GetM.make(selector, header.properties)


def decode_update(buf, header):
    kvs = decode_key_value_list(buf)
    return UpdateM.make(kvs, header.properties)


def decode_delete(buf, header):
    path = buf.get_string()
    return DeleteM.make(path, header.properties)


def decode_sub(buf, header):
    selector = Selector(buf.get_string())
    return SubscribeM.make(selector, header.properties)


def decode_unsub(buf, header):
    subid = buf.get_string()
    return UnsubscribeM.make(subid, header.properties)


def decode_notify(buf, header):
    subid = buf.get_string()
    kvs = decode_key_change_list(buf)
    return NotifyM.make(subid, kvs, header.properties)


def decode_eval(buf, header):
    selector = Selector(buf.get_string())
    return EvalM.make(selector, header)


def decode_reg_eval(buf, header):
    selector = Selector(buf.get_string())
    return RegisterEvalM.make(selector, header.properties)


def decode_unreg_eval(buf, header):
    selector = Selector(buf.get_string())
    return UnregisterEvalM.make(selector, header.properties)


def decode_values(buf, header):
    kvs = decode_key_value_list(buf)
    return ValuesM.make(header, kvs)


def decode_ok(buf, header):
    return OkM.make(header)


def decode_error(buf, header):
    ec = buf.get_vle()
    return ErrorM.make(header, ec)


def decode_message(buf):
    h = decode_header(buf)
    return {
        Message.OK: lambda h: decode_ok(buf, h),
        Message.ERROR: lambda h: decode_error(buf, h),
        Message.PUT: lambda h: decode_update(buf, h),
        Message.GET: lambda h: decode_get(buf, h),
        Message.UPDATE: lambda h: decode_update(buf, h),
        Message.DELETE: lambda h: decode_delete(buf, h),
        Message.SUB: lambda h: decode_sub(buf, h),
        Message.UNSUB: lambda h: decode_unsub(buf, h),
        Message.NOTIFY: lambda h: decode_notify(buf, h),
        Message.EVAL: lambda h: decode_eval(buf, h),
        Message.REG_EVAL: lambda h: decode_reg_eval(buf, h),
        Message.UNREG_EVAL: lambda h: decode_unreg_eval(buf, h),
        Message.VALUES: lambda h: decode_values(buf, h)

    }.get(h.mid)(h)


def encode_message(buf, m):
    {
        Message.LOGIN: lambda m: encode_login(buf, m),
        Message.LOGOUT: lambda m: encode_logout(buf, m),
        Message.WORKSPACE: lambda m: encode_workspace(buf, m),
        Message.PUT: lambda m: encode_put(buf, m),
        Message.GET: lambda m: encode_get(buf, m),
        Message.UPDATE: lambda m: encode_update(buf, m),
        Message.DELETE: lambda m: encode_delete(buf, m),
        Message.VALUES: lambda m: encode_values(buf, m),
        Message.SUB: lambda m: encode_sub(buf, m),
        Message.UNSUB: lambda m: encode_unsub(buf, m),
        Message.NOTIFY: lambda m: encode_notify(buf, m),
        Message.EVAL: lambda m: encode_eval(buf, m),
        Message.REG_EVAL: lambda m: encode_reg_eval(buf, m),
        Message.UNREG_EVAL: lambda m: encode_unreg_eval(buf, m),
        Message.OK: lambda m: encode_ok(buf, m),
        Message.ERROR: lambda m: encode_error(buf, m)
    }.get(m.mid)(m)



# Copyright (c) 2018 ADLINK Technology Inc.
#
# See the NOTICE file(s) distributed with this work for additional
# information regarding copyright ownership.
#
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0, or the Apache License, Version 2.0
# which is available at https://www.apache.org/licenses/LICENSE-2.0.
#
# SPDX-License-Identifier: EPL-2.0 OR Apache-2.0
#
# Contributors: Angelo Corsaro, ADLINK Technology Inc. - Yaks API refactoring

from yaks.yaks import Yaks
from yaks.workspace import Workspace
from yaks.admin import Admin
from yaks.encoding import *
from yaks import exceptions
from yaks.selector import Selector
from yaks.value import Value, Change, ChangeKind
from yaks.path import Path



# Copyright (c) 2018 ADLINK Technology Inc.
#
# See the NOTICE file(s) distributed with this work for additional
# information regarding copyright ownership.
#
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0, or the Apache License, Version 2.0
# which is available at https://www.apache.org/licenses/LICENSE-2.0.
#
# SPDX-License-Identifier: EPL-2.0 OR Apache-2.0
#
# Contributors: Gabriele Baldoni, ADLINK Technology Inc. - Yaks API

import re
import json
from enum import Enum
from yaks.exceptions import ValidationError
from yaks.encoding import Encoding


class ChangeKind(Enum):
    PUT = 'P'
    UPDATE = 'U'
    REMOVE = 'R'


class Value(object):
    def __init__(self, value, encoding=Encoding.RAW, raw_format=""):
        if encoding > Encoding.MAX:
            raise ValueError('Encoding not supported')
        if encoding == Encoding.PROTOBUF:
            raise ValueError('PROTOBUF Encoding not implemented')
        self.encoding = encoding
        if self.encoding == Encoding.JSON:
            if not (isinstance(value, dict) or isinstance(value, str)):
                raise ValidationError("Value is not a valid JSON")
            self.value = json.dumps(value)
        elif self.encoding == Encoding.RAW and isinstance(value, str):
            self.value = value.encode()
        else:
            self.value = value
        self.raw_format = raw_format

    def get_encoding(self):
        return self.encoding

    def get_value(self):
        if self.encoding is Encoding.JSON:
            return json.loads(self.value)
        return self.value

    def __eq__(self, second_value):
        if isinstance(second_value, self.__class__):
            return self.value == second_value.value
        return False

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return self.__str__()


class Change(object):
    def __init__(self, kind, ts, value=None):
        self.set_kind(kind)
        self.ts = ts
        self.value = value

    def set_kind(self, kind):
        kd = {
            'P': ChangeKind.PUT,
            'U': ChangeKind.UPDATE,
            'R': ChangeKind.REMOVE
        }
        if isinstance(kind, ChangeKind):
            self.kind = kind
        else:
            self.kind = kd[kind]

    def get_kind(self):
        return self.kind

    def set_timestamp(self, timestamp):
        self.ts = timestamp

    def get_timestamp(self):
        return self.ts

    def set_value(self, value):
        self.value = value

    def get_value(self):
        return self.value

    def __str__(self):
        return 'Kind: {} TS: {} Value: {}'.format(self.kind,
                                                    self.ts,
                                                    self.value)

    def __repr__(self):
        return self.__str__()



# Copyright (c) 2018 ADLINK Technology Inc.
#
# See the NOTICE file(s) distributed with this work for additional
# information regarding copyright ownership.
#
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0, or the Apache License, Version 2.0
# which is available at https://www.apache.org/licenses/LICENSE-2.0.
#
# SPDX-License-Identifier: EPL-2.0 OR Apache-2.0
#
# Contributors: Angelo Corsaro, ADLINK Technology Inc. - Yaks API refactoring

from yaks.encoding import Encoding, TranscodingFallback
from yaks.message import GetM, PutM, DeleteM, UnsubscribeM, SubscribeM
from yaks.message import RegisterEvalM, UnregisterEvalM, EvalM, Message
from papero.property import *
from yaks.runtime import check_reply_is_ok, check_reply_is_values
from yaks.path import Path
from yaks.selector import Selector
from mvar import MVar
from papero import *


class Workspace(object):
    def __init__(self, runtime, path, wsid):
        self.rt = runtime
        self.path = Path.to_path(path)
        self.wsid = wsid
        self.properties = [Property(Message.WSID, wsid)]
        self.mbox = MVar()
        self.wlbuf = IOBuf()
        self.wbuf = IOBuf()
        self.rbuf = IOBuf()

    def put(self, path, value, quorum=0):
        '''

        The put operation:

        - causes the notification of all subscriptions whose selector matches
            the path parameter, and

        - stores the tuple <path,value> on all storages in YAKS whose selector
            matches the path parameter.

        Notice that the **path** can be absolute or relative to the workspace.

        If a **quorum** is provided then the put will success only if and only
        if a number quorum of independent storages exist that match path.
        If such a set exist, the put operation will complete only after
        the tuple <path,value> has been written on all these storages.

        If no quorum is provided, then no assumptions are made and the **put**
        always succeeds, even if there are currently no matching storage.
        In this case the only effect of this operation will be that of
        triggering matching subscriber, if any exist.

        '''

        path = Path.to_path(path)
        pm = PutM(self.wsid, [(path, value)])
        reply = \
            self.rt.post_message(pm, self.mbox, self.wlbuf, self.wbuf).get()
        return check_reply_is_ok(reply, pm)

    def update(self, path, value, quorum=0):
        '''

        Allows to **put** a delta,
        thus avoiding to distribute the entire value.

        '''

        raise NotImplementedError("Update not yet...")

    def get(self, selector, quorum=0, encoding=Encoding.RAW,
                fallback=TranscodingFallback.KEEP):
        '''

        gets the set of tuples  *<path,value>* available in YAKS
        for which  the path  matches the selector,
        where the selector can be absolute or relative to the workspace.

        If a **quorum** is provided, then **get** will complete succesfully
        if and only if a number **quorum** of independent and complete
        storage set exist.
        Complete storage means a storage that fully covers the selector
        (i.e. any path matching the selector is covered by the storage).
        This ensures that if there is a  *{ <path,value> }* stored in YAKS
        for which the *path* matches the selector **s**,
        then there are at least **quorum** idependent copies of this element
        stored in YAKS.
        Of these **quorum** idependent copies,
        the one returned to the application is the most recent version.

        If no quorum is provided (notice this is the default behaviour) then
        the [get] will succeed even if there isn't a set of storages that
        fully covers the selector.
        I.e. storages that partially cover the selector will also reply.

        The **encoding**  allows an application to request values to be
        encoded in a specific format.


        If no encoding is provided (this is the default behaviour) then YAKS
        will not try to perform any transcoding and will
        return matching values in the encoding in which they are stored.

        The **fallback** controls what happens for those values that cannot be
        transcoded into the desired encoding, the available options are:

        - Fail: the **get** fails if some value cannot be transcoded.
        - Drop: values that cannot be transcoded are dropped.
        - Keep: values that cannot be transcoded are kept with their original
            encoding and left for the application to deal with.

        '''

        s = Selector.to_selector(selector)
        gm = GetM(self.wsid, s)
        reply = \
             self.rt.post_message(gm, self.mbox, self.wlbuf, self.wbuf).get()
        if check_reply_is_values(reply, gm):
            return reply.kvs
        else:
            raise "Get received an invalid reply"
        return []

    def remove(self, path, quorum=0):
        '''

        Removes from all  Yaks's storages the tuples having the given **path**.
        The **path** can be absolute or relative to the workspace.
        If a **quorum** is provided, then the *remove* will
        complete only after having successfully removed the tuple
        from **quorum** storages.

        '''

        path = Path.to_path(path)
        rm = DeleteM(self.wsid, path)
        reply = \
             self.rt.post_message(rm, self.mbox, self.wlbuf, self.wbuf).get()
        return check_reply_is_ok(reply, rm)

    def subscribe(self, selector, listener=None):
        '''

        Registers a subscription to tuples whose path matches the **selector**.

        A subscription identifier is returned.
        The **selector** can be absolute or relative to the workspace.
        If specified,  the **listener callback will be called for each **put**
        and **update** on tuples whose
        path matches the subscription **selector**
        listener should expect a list of (Path, Changes)

        '''

        s = Selector.to_selector(selector)
        sm = SubscribeM(self.wsid, s)
        reply = \
             self.rt.post_message(sm, self.mbox, self.wlbuf, self.wbuf).get()
        if check_reply_is_ok(reply, sm):
            subid = find_property(Message.SUBID, reply.properties)
            if listener is not None:
                self.rt.add_listener(subid, listener)
            return subid
        else:
            raise "Subscribe received an invalid reply"

    def unsubscribe(self, subscription_id):
        '''

        Unregisters a previous subscription with the identifier **subid**

        '''
        um = UnsubscribeM(self.wsid, subscription_id)
        reply = \
            self.rt.post_message(um, self.mbox, self.wlbuf, self.wbuf).get()
        if check_reply_is_ok(reply, um):
            self.rt.remove_listener(subscription_id)
            return True
        else:
            raise "Unsubscribe received an invalid reply"

    def register_eval(self, path, callback):
        '''

        Registers an evaluation function **eval** under the provided **path**.
        The **path** can be absolute or relative to the workspace.

        '''

        path = Path.to_path(path)
        rem = RegisterEvalM(self.wsid, path)
        reply = \
             self.rt.post_message(rem, self.mbox, self.wlbuf, self.wbuf).get()
        if check_reply_is_ok(reply, rem):
            self.rt.add_eval_callback(path, callback)
            return True
        else:
            raise "Register_eval received an invalid reply"

    def unregister_eval(self, path):
        '''

        Unregisters an previously registered evaluation function under
        the give [path].
        The [path] can be absolute or relative to the workspace.

        '''

        path = Path.to_path(path)
        uem = UnregisterEvalM(self.wsid, path)
        reply = \
            self.rt.post_message(uem, self.mbox, self.wlbuf, self.wbuf).get()
        if check_reply_is_ok(reply, uem):
            self.rt.remove_eval_callback(path)
            return True
        else:
            raise "Unregister_eval received an invalid reply"

    def eval(self, selector, multiplicity=1, encoding=Encoding.RAW,
             fallback=TranscodingFallback.KEEP):
        '''

        Requests the evaluation of registered evals whose registration
        **path** matches the given **selector**.

        If several evaluation function are registerd with the same path
        (by different Yaks clients), then Yaks will call N functions
        where N=[multiplicity] (default value is 1).
        Note that in such case, the returned *{ <path,value> }*
        will contain N time each matching path with the different
        values returned by each evaluation.
        The **encoding** indicates the expected encoding of the resulting
        values. If the original values have a different encoding,
        Yaks will try to transcode them into the expected encoding.
        By default, if no encoding is specified, the values are
        returned with their original encoding.
        The **fallback** indicates the action that YAKS
        will perform if the transcoding of a value fails.

        '''

        s = Selector.to_selector(selector)
        em = EvalM(self.wsid, s)
        reply = \
            self.rt.post_message(em, self.mbox, self.wlbuf, self.wbuf).get()
        if check_reply_is_values(reply, em):
            return reply.kvs
        else:
            raise "Get received an invalid reply"
        return []



# Copyright (c) 2018 ADLINK Technology Inc.
#
# See the NOTICE file(s) distributed with this work for additional
# information regarding copyright ownership.
#
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0, or the Apache License, Version 2.0
# which is available at https://www.apache.org/licenses/LICENSE-2.0.
#
# SPDX-License-Identifier: EPL-2.0 OR Apache-2.0
#
# Contributors: Angelo Corsaro, ADLINK Technology Inc. - Yaks API refactoring

import socket
import uuid
import os
import threading
import logging
import sys
import traceback
from yaks.logger import APILogger
from papero import *
from mvar import MVar
from yaks.codec import decode_message, encode_message
from yaks.message import Message, ErrorM, LogoutM, ValuesM


def get_frame_len(sock, buf):
    buf.clear()
    v = 0xff
    while v > 0x7f:
        b = sock.recv(1)
        v = byte_to_int(b)
        buf.put(v)

    return buf.get_vle()


def recv_msg(sock, lbuf):
    lbuf.clear()
    try:
        flen = get_frame_len(sock, lbuf)

        bs = sock.recv(flen)
        n = len(bs)
        while n < flen:
            m = flen - n
            cs = sock.recv(m)
            n = n + len(cs)
            bs = bs + cs

        rbuf = IOBuf.from_bytes(bs)
        m = decode_message(rbuf)
        return m
    except OSError:
        m = ErrorM(0)
        return m


def send_msg(sock, msg, buf, lbuf):
    buf.clear()
    lbuf.clear()
    encode_message(buf, msg)
    length = buf.write_pos
    lbuf.put_vle(length)
    sock.sendall(lbuf.get_raw_bytes())
    sock.sendall(buf.get_raw_bytes())


def get_log_level():
    log_level = logging.ERROR
    i = 0
    for a in sys.argv:
        if a.startswith('--log='):
            _, _, lvl = a.partition('=')
            log_level = getattr(logging, lvl.upper())
        elif a == '-l':
            log_level = getattr(logging, sys.argv[i + 1].upper())
        i += 1
    return log_level


def check_reply_is_ok(reply, msg):
    if reply.mid == Message.OK and msg.corr_id == reply.corr_id:
        return True
    elif reply.mid == Message.ERROR:
        raise RuntimeError(
            'Yaks refused connection because of {}'.format(reply.error_code))
    else:
        raise RuntimeError('Yaks replied with unexpected message')


def check_reply_is_values(reply, msg):
    if reply.mid == Message.VALUES and msg.corr_id == reply.corr_id:
        return True
    elif reply.mid == Message.ERROR:
        raise RuntimeError(
            'Yaks refused connection because of {}'.format(reply.error_code))
    else:
        raise RuntimeError('Yaks replied with unexpected message')


class Runtime(threading.Thread):
    DEFAULT_TIMEOUT = 5

    def __init__(self, sock, locator, on_close):
        threading.Thread.__init__(self)
        self.logger = APILogger(get_log_level(), True)
        self.daemon = True
        self.connected = True
        self.posted_messages = {}
        self.sock = sock
        self.running = True
        self.on_close = on_close
        self.listeners = {}
        self.eval_callbacks = {}
        self.putMbox = MVar()
        self.rtMbox = MVar()
        self.evalMBox = MVar()
        self.tss = {}
        self.wlbuf = IOBuf()
        self.wbuf = IOBuf()
        self.rlbuf = IOBuf()
        self.wlbuf = IOBuf()

    def close(self):
        self.post_message(LogoutM(), self.rtMbox, self.wlbuf, self.wbuf).get()
        self.on_close(self)
        self.running = False
        self.sock.close()

    def post_message(self, msg, mbox, wlbuf, wbuf):
        self.posted_messages.update({msg.corr_id: (mbox, [])})
        send_msg(self.sock, msg, wbuf, wlbuf)
        return mbox

    def add_listener(self, subid, callback):
        self.listeners.update({subid: callback})

    def remove_listener(self, subid):
        if subid in self.listeners.keys():
            self.listeners.pop(subid)

    def add_eval_callback(self, path, callback):
        self.eval_callbacks.update({path: callback})

    def remove_eval_callback(self, path):
        if path in self.eval_callbacks.keys():
            self.eval_callbacks.pop(path)

    def notify_listeners(self, m):
        subid = m.subid
        listener = self.listeners.get(subid)
        if listener:
            l_th = threading.Thread(target=listener, args=(m.kvs,))
            l_th.start()

    def execute_eval(self, m):
        selector = m.selector
        for path in self.eval_callbacks:
            if selector.is_prefixed_by_path(path):
                cb = self.eval_callbacks.get(path)
                p = selector.get_path()
                args = selector.dict_from_properties()
                # TODO: should be called in another thread

                def eval_cb_adaptor(path, p, args, cid):
                    try:
                        kvs = [(path, cb(p, **args))]
                        vm = ValuesM(kvs)
                        vm.corr_id = cid
                        reply = self.post_message(vm, self.evalMBox,
                                                        self.wlbuf,
                                                        self.wbuf).get()
                        if not check_reply_is_ok(reply, vm):
                            raise ValueError('YAKS error on EVAL')
                    except (Exception, RuntimeError):
                        self.post_message(ErrorM.make(cid, ErrorM.BAD_REQUEST),
                                                        self.evalMBox,
                                                        self.wlbuf,
                                                        self.wbuf)

                eval_th = threading.Thread(target=eval_cb_adaptor,
                                           args=(path, p, args, m.corr_id))
                eval_th.start()

    def consolidate_reply(self, ms):
        ckvs = []
        for m in ms:
            ckvs = ckvs + m.kvs

        m = ms[len(ms) - 1]
        m.kvs = ckvs
        return m

    def handle_reply(self, m):
        map_value = self.posted_messages.get(m.corr_id)
        if map_value is None:
            self.logger.warning('handle_reply()',
                                '>> Received not matching message {}'
                                .format(m.corr_id))
            return
        self.logger.warning('handle_reply()',
                                '>> Received matching message {}'
                                .format(m.corr_id))
        (mvar, partial) = map_value
        if mvar is not None:
            if m.mid == Message.VALUES:
                if m.is_complete():
                    self.logger.warning('handle_reply()',
                                        '>> Complete message received {}'
                                        .format(m.corr_id))
                    if len(partial) > 0:
                        partial.append(m)
                        m = self.consolidate_reply(partial)
                    mvar.put(m)
                    self.posted_messages.pop(m.corr_id)
                else:
                    self.logger.warning('handle_reply()',
                                        '>> Partial message received {}'
                                        .format(m.corr_id))
                    partial.append(m)
                    self.posted_messages.update({m.corr_id: (mvar, partial)})

            else:
                mvar.put(m)
                self.posted_messages.pop(m.corr_id)

    def handle_unexpected_message(self, m):
        self.logger.warning('handle_unexpected_message()',
                            '>> Received unexpected message with id {}\
                             -- ignoring'
                            .format(m.mid))

    def run(self):
        try:
            while self.running:
                m = recv_msg(self.sock, self.rlbuf)
                self.logger.debug('run()',
                                  '>> Received msg with id: {}'.format(m.mid))
                {
                    Message.NOTIFY: lambda m: self.notify_listeners(m),
                    Message.EVAL: lambda m: self.execute_eval(m),
                    Message.OK: lambda m: self.handle_reply(m),
                    Message.VALUES: lambda m: self.handle_reply(m),
                    Message.ERROR: lambda m: self.handle_reply(m)
                }.get(m.mid, lambda m: self.handle_unexpected_message(m))(m)

        except Exception as e:
            traceback.print_exc()
            self.logger.debug('run()',
                              'Terminating the runloop because of {}'
                              .format(e))



# Copyright (c) 2018 ADLINK Technology Inc.
#
# See the NOTICE file(s) distributed with this work for additional
# information regarding copyright ownership.
#
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0, or the Apache License, Version 2.0
# which is available at https://www.apache.org/licenses/LICENSE-2.0.
#
# SPDX-License-Identifier: EPL-2.0 OR Apache-2.0
#
# Contributors: Gabriele Baldoni, ADLINK Technology Inc. - Yaks API

import re
from yaks.exceptions import ValidationError


class Path(object):
    def __init__(self, path):
        self.__path_regex = re.compile('^[^?#*]+$')
        if not self.is_valid(path):
            raise ValidationError("{} is not a valid Path".format(path))
        self.path = path

    @staticmethod
    def to_path(p):
        if isinstance(p, Path):
            return p
        else:
            return Path(p)

    def is_valid(self, path):
        return self.__path_regex.match(path) is not None \
            and not path.startswith('//')

    def is_absolute(self):
        if self.path.startswith('/'):
            return True
        return False

    def is_prefix(self, prefix):
        return self.path.startswith(prefix)

    def remove_prefix(self, prefix):
        if self.is_prefix(prefix):
            self.path = self.path[len(prefix):]

    def to_string(self):
        return self.path

    def __len__(self):
        return len(self.path)

    def __eq__(self, second_path):
        if isinstance(second_path, self.__class__):
            return self.path == second_path.path
        return False

    def __str__(self):
        return self.path

    def __repr__(self):
        return self.__str__()

    def __hash__(self):
        return self.path.__hash__()



# Copyright (c) 2018 ADLINK Technology Inc.
#
# See the NOTICE file(s) distributed with this work for additional
# information regarding copyright ownership.
#
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0, or the Apache License, Version 2.0
# which is available at https://www.apache.org/licenses/LICENSE-2.0.
#
# SPDX-License-Identifier: EPL-2.0 OR Apache-2.0
#
# Contributors: Gabriele Baldoni, ADLINK Technology Inc. - Yaks API

# Encoding

from enum import Enum

# TODO: This should be changed in enum


class Encoding(object):
    RAW = 0x01
    STRING = 0x02
    JSON = 0x03
    PROTOBUF = 0x04
    SQL = 0x05
    PROPERTY = 0x6
    # The following are both invalid encoding,
    # only numbers in this range are valid
    MIN = 0x00
    MAX = 0xff


class TranscodingFallback(Enum):
    FAIL = 0x01
    DROP = 0x02
    KEEP = 0x03



# Copyright (c) 2018 ADLINK Technology Inc.
#
# See the NOTICE file(s) distributed with this work for additional
# information regarding copyright ownership.
#
# This program and the accompanying materials are made available under the
# terms of the LGPL 2.1 which is available at
# https://github.com/atolab/yaks/blob/master/LICENSE
# Contributors: Gabriele Baldoni, ADLINK Technology Inc. - API

#!/usr/bin/env python3

from setuptools import setup
import os


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name='yaks',
    version='0.2.7',
    author='ADLINK Advance Technology Office',
    description='Python API to access the YAKS service',
    long_description=read('README.md'),
    packages=['yaks'],
    url='https://github.com/atolab/yaks-python',
    authon_email='gabriele.baldoni@adlinktech.com',
    install_requires=['hexdump', 'mvar', 'papero==0.2.7'],
    license='Apache 2.O or EPL 2.0',
    classifiers=[
          'Development Status :: 2 - Pre-Alpha',
          'Intended Audience :: Developers',
          'Intended Audience :: Telecommunications Industry',
          'License :: OSI Approved :: Apache Software License',
          'License :: OSI Approved :: Eclipse Public License 2.0 (EPL-2.0)',
          'Programming Language :: Python :: 3',
        #   'Programming Language :: Python :: 2'
    ],
    include_package_data=True
)



# Copyright (c) 2018 ADLINK Technology Inc.
#
# See the NOTICE file(s) distributed with this work for additional
# information regarding copyright ownership.
#
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0, or the Apache License, Version 2.0
# which is available at https://www.apache.org/licenses/LICENSE-2.0.
#
# SPDX-License-Identifier: EPL-2.0 OR Apache-2.0
#
# Contributors: Angelo Corsaro, ADLINK Technology Inc. - Yaks API refactoring


from yaks.runtime import check_reply_is_ok, socket, send_msg, Runtime, recv_msg
from yaks.workspace import Workspace
from yaks.message import LoginM, LogoutM, WorkspaceM, Message
from papero import find_property
from yaks.admin import *
from papero import IOBuf
from mvar import MVar
import threading


class Yaks(object):
    DEFAULT_PORT = 7887

    def __init__(self, rt):
        self.rt = rt
        self.mbox = MVar()
        self.wbuf = IOBuf()
        self.wlbuf = IOBuf()

    @staticmethod
    def login(locator, properties=None, on_close=lambda z: z, lease=0):
        '''

        Establish a session with the Yaks instance reachable through the
        provided *locator*.

        Valid format for the locator are valid  IP addresses as well
        as the combination IP:PORT.

        '''
        addr, _, p = locator.partition(':')
        if p == '':
            port = Yaks.DEFAULT_PORT
        else:
            port = int(p)

        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        sock.setblocking(1)
        sock.connect((addr, port))

        wbuf = IOBuf()
        lbuf = IOBuf()
        login = LoginM(properties)
        send_msg(sock, login, wbuf, lbuf)
        m = recv_msg(sock, lbuf)
        # y = None
        if check_reply_is_ok(m, login):
            rt = Runtime(sock, locator, on_close)
        rt.start()
        return Yaks(rt)

    def workspace(self, path):
        '''

        Creates a workspace relative to the provided **path**.
        Any *put* or *get* operation with relative paths on this workspace
        will be prepended with the workspace *path*.

        '''

        wsm = WorkspaceM(path)
        reply =  \
            self.rt.post_message(wsm, self.mbox, self.wlbuf, self.wbuf).get()
        ws = None
        wsid = None
        if check_reply_is_ok(reply, wsm):
            wsid = find_property(Message.WSID, reply.properties)
            if wsid is None:
                raise "Workspace id was not provided by YAKS"
            else:
                ws = Workspace(self.rt, path, wsid)
        return ws

    def logout(self):
        '''
        Terminates this session.
        '''
        self.rt.close()
        # lom = LogoutM()
        # reply = self.rt.post_message(lom).get()
        # if check_reply_is_ok(reply, lom):

    def admin(self):
        '''

        Creates an admin workspace that provides helper operations to
        administer Yaks.

        '''
        return Admin(self.workspace(
            '/{}/{}'.format(Admin.PREFIX, Admin.MY_YAKS)))



# Copyright (c) 2018 ADLINK Technology Inc.
#
# See the NOTICE file(s) distributed with this work for additional
# information regarding copyright ownership.
#
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0, or the Apache License, Version 2.0
# which is available at https://www.apache.org/licenses/LICENSE-2.0.
#
# SPDX-License-Identifier: EPL-2.0 OR Apache-2.0
#
# Contributors: Gabriele Baldoni, ADLINK Technology Inc. - Yaks API

import logging
import logging.handlers
import sys
import os


class APILogger:
    class __SingletonLogger:
        def __init__(self, level, debug_flag):

            log_format = '[%(asctime)s] - [%(levelname)s] > %(message)s'

            self.logger = logging.getLogger('is.yaks.python.api')

            self.logger.setLevel(level)
            formatter = logging.Formatter(log_format)
            if not debug_flag:
                platform = sys.platform
                if platform == 'linux':
                    handler = logging.handlers.SysLogHandler('/dev/log')
                elif platform == 'darwin':
                    handler = logging.handlers.SysLogHandler('/var/run/syslog')
                elif platform in ['windows', 'Windows', 'win32']:
                    handler = logging.handlers.SysLogHandler()
            else:
                handler = logging.StreamHandler(sys.stdout)
            handler.setFormatter(formatter)
            self.logger.addHandler(handler)

        def info(self, caller, message):
            self.logger.info('< {} > {}'.format(caller, message))

        def warning(self, caller, message):
            self.logger.warning('< {} > {}'.format(caller, message))

        def error(self, caller, message):
            self.logger.error('< {} > {}'.format(caller, message))

        def debug(self, caller, message):
            self.logger.debug('< {} > {}'.format(caller, message))

    instance = None
    enabled = True

    def __init__(self, level, debug_flag):

        if not APILogger.instance:
            APILogger.instance = \
                APILogger.__SingletonLogger(level, debug_flag)

    def enable(self):
        self.enabled = True

    def disable(self):
        self.enabled = False

    def info(self, caller, message):
        if self.enabled:
            self.instance.info(caller, message)

    def warning(self, caller, message):
        if self.enabled:
            self.instance.warning(caller, message)

    def error(self, caller, message):
        if self.enabled:
            self.instance.error(caller, message)

    def debug(self, caller, message):
        if self.enabled:
            self.instance.debug(caller, message)
